@@ -87,8 +87,8 @@ def _param_to_openapi(param):
         type_ = 'boolean'
     elif isinstance(san, BoolSanitizer):
         type_ = 'boolean'
-        #type_ = 'string'
-        #definition['examples'] = {choice: {'value': choice, 'summary': choice} for choice in san.choices}
+        # type_ = 'string'
+        # definition['examples'] = {choice: {'value': choice, 'summary': choice} for choice in san.choices}
     elif isinstance(san, ChoicesSanitizer):
         type_ = 'string'
         definition['examples'] = {choice: {'value': choice, 'summary': choice} for choice in san.choices}
@@ -385,6 +385,8 @@ class _OpenAPIBase:
             },
             'objects.get.query.hidden': {
             },
+            'objects.get.query.opened': {
+            },
             'objects.get.query.superordinate': {
             },
             'objects.get.query.properties': {
@@ -397,6 +399,8 @@ class _OpenAPIBase:
             },
             'objects.get.query.by': {
             },
+            'object.get.query.properties': {
+            },
             'object.delete.query.cleanup': {
             },
             'object.delete.query.recursive': {
@@ -507,7 +511,7 @@ class _OpenAPIBase:
                 }),
             },
             "PUTObjectCreated": {  # 201
-                "description": "Created: The object did not exist and has been created. Deprecated: a move operation started, expect 202 in the future!",
+                "description": "Created: The object did not exist and has been created.",
                 "content": content_schema({
                     "type": "object",
                             "properties": {
@@ -520,7 +524,7 @@ class _OpenAPIBase:
                     'Last-Modified': {'$ref': '#/components/headers/Last-Modified'},
                 }),
             },
-            'MoveStarted': {  # 202 (actually still 201)
+            'MoveStarted': {  # 202
                 "description": "Accepted: asynchronous move or rename operation started.",
                 'headers': global_response_headers({
                     'Location': {'$ref': '#/components/headers/Location'},
@@ -789,7 +793,10 @@ class _OpenAPIBase:
             openapi_paths[object_path]["get"] = {
                 "operationId": f"udm:{name}/object",
                 "summary": docstring('object', 'get', module),
-                "parameters": [] + global_parameters,
+                "parameters": [
+                    {'$ref': '#/components/parameters/object.get.query.properties'},
+                    # {'$ref': '#/components/parameters/object.get.query.copy'},
+                ] + global_parameters,
                 "responses": global_responses({
                     "200": {
                         '$ref': '#/components/responses/object.%s.get.response.success' % (_openapi_quote(model_name),),
